@@ -3,23 +3,20 @@
 files according to rfc2445.
 
 These are the defined components.
-
 """
-from __future__ import absolute_import
-import pytz
 from datetime import datetime
-from .parser_tools import data_encode
-from .caselessdict import CaselessDict
-from .parser import (
-    Contentlines,
-    Contentline,
-    Parameters,
-    q_split,
-    q_join,
-)
-from .prop import TypesFactory
-from .prop import vText, vDDDLists
-from .parser_tools import DEFAULT_ENCODING
+from icalendar.caselessdict import CaselessDict
+from icalendar.parser import Contentline
+from icalendar.parser import Contentlines
+from icalendar.parser import Parameters
+from icalendar.parser import q_join
+from icalendar.parser import q_split
+from icalendar.parser_tools import DEFAULT_ENCODING
+from icalendar.parser_tools import data_encode
+from icalendar.prop import TypesFactory
+from icalendar.prop import vText, vDDDLists
+
+import pytz
 
 
 ######################################
@@ -28,13 +25,12 @@ from .parser_tools import DEFAULT_ENCODING
 class ComponentFactory(CaselessDict):
     """All components defined in rfc 2445 are registered in this factory class.
     To get a component you can use it like this.
-
     """
 
     def __init__(self, *args, **kwargs):
         """Set keys to upper for initial dict.
         """
-        CaselessDict.__init__(self, *args, **kwargs)
+        super(ComponentFactory, self).__init__(*args, **kwargs)
         self['VEVENT'] = Event
         self['VTODO'] = Todo
         self['VJOURNAL'] = Journal
@@ -48,9 +44,11 @@ class ComponentFactory(CaselessDict):
 
 # These Properties have multiple property values inlined in one propertyline
 # seperated by comma. Use CaselessDict as simple caseless set.
-INLINE = CaselessDict(
-    [(cat, 1) for cat in ('CATEGORIES', 'RESOURCES', 'FREEBUSY')]
-)
+INLINE = CaselessDict({
+    'CATEGORIES': 1,
+    'RESOURCES': 1,
+    'FREEBUSY': 1,
+})
 
 _marker = []
 
@@ -59,15 +57,15 @@ class Component(CaselessDict):
     """Component is the base object for calendar, Event and the other
     components defined in RFC 2445. normally you will not use this class
     directy, but rather one of the subclasses.
-
     """
 
-    name = ''       # must be defined in each component
-    required = ()   # These properties are required
-    singletons = () # These properties must only appear once
-    multiple = ()   # may occur more than once
-    exclusive = ()  # These properties are mutually exclusive
-    inclusive = ()  # if any occurs the other(s) MUST occur ('duration', 'repeat')
+    name = ''        # must be defined in each component
+    required = ()    # These properties are required
+    singletons = ()  # These properties must only appear once
+    multiple = ()    # may occur more than once
+    exclusive = ()   # These properties are mutually exclusive
+    inclusive = ()   # if any occurs the other(s) MUST occur
+                     # ('duration', 'repeat')
     ignore_exceptions = False   # if True, and we cannot parse this
                                 # component, we will silently ignore
                                 # it, rather than let the exception
@@ -76,12 +74,12 @@ class Component(CaselessDict):
 
     def __init__(self, *args, **kwargs):
         """Set keys to upper for initial dict.
-
         """
-        CaselessDict.__init__(self, *args, **kwargs)
+        super(Component, self).__init__(*args, **kwargs)
         # set parameters here for properties that use non-default values
-        self.subcomponents = [] # Components can be nested.
-        self.is_broken = False  # True iff we ignored an exception while parsing a property
+        self.subcomponents = []  # Components can be nested.
+        self.is_broken = False  # True if we ignored an exception while
+                                # parsing a property
 
     #def is_compliant(self, name):
     #    """Returns True is the given property name is compliant with the
@@ -93,38 +91,67 @@ class Component(CaselessDict):
     #    """
     #    return name in not_compliant
 
-
     #############################
     # handling of property values
 
-    def _encode(self, name, value, cond=1):
-        """Conditional convertion of values.
+    def _encode(self, name, value, parameters=None, encode=1):
+        """Encode values to icalendar property values.
 
+        :param name: Name of the property.
+        :type name: string
+
+        :param value: Value of the property. Either of a basic Python type of
+                      any of the icalendar's own property types.
+        :type value: Python native type or icalendar property type.
+
+        :param parameters: Property parameter dictionary for the value. Only
+                           available, if encode is set to True.
+        :type parameters: Dictionary
+
+        :param encode: True, if the value should be encoded to one of
+                       icalendar's own property types (Fallback is "vText")
+                       or False, if not.
+        :type encode: Boolean
+
+        :returns: icalendar property value
         """
-        if not cond:
+        if not encode:
             return value
-        if type(value) in types_factory.all_types:
+        if isinstance(value, types_factory.all_types):
             # Don't encode already encoded values.
             return value
         klass = types_factory.for_property(name)
         obj = klass(value)
-        if hasattr(value, 'params') and len(value.params.keys()) > 0:
-            # TODO: How can a python native value have params?
-            obj.params = value.params
+        if parameters:
+            if isinstance(parameters, dict):
+                params = Parameters()
+                for key, item in parameters.items():
+                    params[key] = item
+                parameters = params
+            assert isinstance(parameters, Parameters)
+            obj.params = parameters
         return obj
 
-    def set(self, name, value, encode=1):
-        if encode and isinstance(value, list) \
-                and name.lower() not in ['rdate', 'exdate']:
-            # Individually convert each value to an ical type except rdate and
-            # exdate, where lists of dates might be passed to vDDDLists.
-            self[name] = [self._encode(name, v, encode) for v in value]
-        else:
-            self[name] = self._encode(name, value, encode)
-
-    def add(self, name, value, encode=1):
+    def add(self, name, value, parameters=None, encode=1):
         """Add a property.
 
+        :param name: Name of the property.
+        :type name: string
+
+        :param value: Value of the property. Either of a basic Python type of
+                      any of the icalendar's own property types.
+        :type value: Python native type or icalendar property type.
+
+        :param parameters: Property parameter dictionary for the value. Only
+                           available, if encode is set to True.
+        :type parameters: Dictionary
+
+        :param encode: True, if the value should be encoded to one of
+                       icalendar's own property types (Fallback is "vText")
+                       or False, if not.
+        :type encode: Boolean
+
+        :returns: None
         """
         if isinstance(value, datetime) and\
                 name.lower() in ('dtstamp', 'created', 'last-modified'):
@@ -135,20 +162,31 @@ class Component(CaselessDict):
                 # assume UTC for naive datetime instances
                 value = pytz.utc.localize(value)
 
-        # If property already exists, append it. Otherwise create and set it.
-        if name in self:
-            oldval = self[name]
-            value = self._encode(name, value, encode)
-            if isinstance(oldval, list):
-                oldval.append(value)
-            else:
-                self.set(name, [oldval, value], encode=0)
+        # encode value
+        if encode and isinstance(value, list) \
+                and name.lower() not in ['rdate', 'exdate']:
+            # Individually convert each value to an ical type except rdate and
+            # exdate, where lists of dates might be passed to vDDDLists.
+            value = [self._encode(name, v, parameters, encode) for v in value]
         else:
-            self.set(name, value, encode)
+            value = self._encode(name, value, parameters, encode)
+
+        # set value
+        if name in self:
+            # If property already exists, append it.
+            oldval = self[name]
+            if isinstance(oldval, list):
+                if isinstance(value, list):
+                    value = oldval + value
+                else:
+                    oldval.append(value)
+                    value = oldval
+            else:
+                value = [oldval, value]
+        self[name] = value
 
     def _decode(self, name, value):
         """Internal for decoding property values.
-
         """
 
         # TODO: Currently the decoded method calls the icalendar.prop instances
@@ -167,7 +205,6 @@ class Component(CaselessDict):
 
     def decoded(self, name, default=_marker):
         """Returns decoded value of property.
-
         """
         # XXX: fail. what's this function supposed to do in the end?
         # -rnix
@@ -189,7 +226,6 @@ class Component(CaselessDict):
 
     def get_inline(self, name, decode=1):
         """Returns a list of values (split on comma).
-
         """
         vals = [v.strip('" ') for v in q_split(self[name])]
         if decode:
@@ -199,10 +235,9 @@ class Component(CaselessDict):
     def set_inline(self, name, values, encode=1):
         """Converts a list of values into comma seperated string and sets value
         to that.
-
         """
         if encode:
-            values = [self._encode(name, value, 1) for value in values]
+            values = [self._encode(name, value, encode=1) for value in values]
         self[name] = types_factory['inline'](q_join(values))
 
     #########################
@@ -210,13 +245,11 @@ class Component(CaselessDict):
 
     def add_component(self, component):
         """Add a subcomponent to this component.
-
         """
         self.subcomponents.append(component)
 
     def _walk(self, name):
         """Walk to given component.
-
         """
         result = []
         if name is None or self.name == name:
@@ -228,7 +261,6 @@ class Component(CaselessDict):
     def walk(self, name=None):
         """Recursively traverses component and subcomponents. Returns sequence
         of same. If name is passed, only components with name will be returned.
-
         """
         if not name is None:
             name = name.upper()
@@ -237,14 +269,17 @@ class Component(CaselessDict):
     #####################
     # Generation
 
-    def property_items(self, recursive=True):
+    def property_items(self, recursive=True, sorted=True):
         """Returns properties in this component and subcomponents as:
         [(name, value), ...]
-
         """
         vText = types_factory['text']
         properties = [('BEGIN', vText(self.name).to_ical())]
-        property_names = self.sorted_keys()
+        if sorted:
+            property_names = self.sorted_keys()
+        else:
+            property_names = self.keys()
+
         for name in property_names:
             values = self[name]
             if isinstance(values, list):
@@ -256,21 +291,32 @@ class Component(CaselessDict):
         if recursive:
             # recursion is fun!
             for subcomponent in self.subcomponents:
-                properties += subcomponent.property_items()
+                properties += subcomponent.property_items(sorted=sorted)
         properties.append(('END', vText(self.name).to_ical()))
         return properties
 
     @classmethod
     def from_ical(cls, st, multiple=False):
         """Populates the component recursively from a string.
-
         """
-        stack = [] # a stack of components
+        stack = []  # a stack of components
         comps = []
-        for line in Contentlines.from_ical(st): # raw parsing
+        for line in Contentlines.from_ical(st):  # raw parsing
             if not line:
                 continue
-            name, params, vals = line.parts()
+
+            try:
+                name, params, vals = line.parts()
+            except ValueError:
+                # if unable to parse a line within a component
+                # that ignores exceptions, mark the component
+                # as broken and skip the line. otherwise raise.
+                component = stack[-1] if stack else None
+                if not component or not component.ignore_exceptions:
+                    raise
+                component.is_broken = True
+                continue
+
             uname = name.upper()
             # check for start of component
             if uname == 'BEGIN':
@@ -279,7 +325,7 @@ class Component(CaselessDict):
                 c_name = vals.upper()
                 c_class = component_factory.get(c_name, cls)
                 component = c_class()
-                if not getattr(component, 'name', ''): # undefined components
+                if not getattr(component, 'name', ''):  # undefined components
                     component.name = c_name
                 stack.append(component)
             # check for end of event
@@ -287,7 +333,7 @@ class Component(CaselessDict):
                 # we are done adding properties to this component
                 # so pop it from the stack and add it to the new top.
                 component = stack.pop()
-                if not stack: # we are at the end
+                if not stack:  # we are at the end
                     comps.append(component)
                 else:
                     if not component.is_broken:
@@ -296,9 +342,10 @@ class Component(CaselessDict):
             else:
                 factory = types_factory.for_property(name)
                 component = stack[-1]
+                datetime_names = ('DTSTART', 'DTEND', 'RECURRENCE-ID', 'DUE',
+                                  'FREEBUSY', 'RDATE', 'EXDATE')
                 try:
-                    if name in ('DTSTART', 'DTEND', 'RECURRENCE-ID')\
-                            and 'TZID' in params: # TODO: add DUE, FREEBUSY, RDATE, EXDATE ..
+                    if name in datetime_names and 'TZID' in params:
                         vals = factory(factory.from_ical(vals, params['TZID']))
                     else:
                         vals = factory(factory.from_ical(vals))
@@ -324,24 +371,29 @@ class Component(CaselessDict):
     def __repr__(self):
         return '%s(%s)' % (self.name, data_encode(self))
 
-    def content_line(self, name, value):
+    def content_line(self, name, value, sorted=True):
         """Returns property as content line.
         """
         params = getattr(value, 'params', Parameters())
-        return Contentline.from_parts(name, params, value)
+        return Contentline.from_parts(name, params, value, sorted=sorted)
 
-    def content_lines(self):
+    def content_lines(self, sorted=True):
         """Converts the Component and subcomponents into content lines.
         """
         contentlines = Contentlines()
-        for name, value in self.property_items():
-            cl = self.content_line(name, value)
+        for name, value in self.property_items(sorted=sorted):
+            cl = self.content_line(name, value, sorted=sorted)
             contentlines.append(cl)
-        contentlines.append('') # remember the empty string in the end
+        contentlines.append('')  # remember the empty string in the end
         return contentlines
 
-    def to_ical(self):
-        content_lines = self.content_lines()
+    def to_ical(self, sorted=True):
+        '''
+        :param sorted: Whether parameters and properties should be
+        lexicographically sorted.
+        '''
+
+        content_lines = self.content_lines(sorted=sorted)
         return content_lines.to_ical()
 
 
@@ -451,7 +503,6 @@ class Alarm(Component):
 
 class Calendar(Component):
     """This is the base object for an iCalendar file.
-
     """
     name = 'VCALENDAR'
     canonical_order = ('VERSION', 'PRODID', 'CALSCALE', 'METHOD',)
